import h5py
import numpy as np
import ismrmrd

from .constants import *

# For Python 2.7 ctypes bug
import warnings

encoding_counters_dtype = np.dtype(
       [('kspace_encode_step_1', '<u2'),
        ('kspace_encode_step_2', '<u2'),
        ('average', '<u2'),
        ('slice', '<u2'),
        ('contrast', '<u2'),
        ('phase', '<u2'),
        ('repetition', '<u2'),
        ('set', '<u2'),
        ('segment', '<u2'),
        ('user', '<u2', (8,))])

acquisition_header_dtype = np.dtype(
     [('version', '<u2'),
      ('flags', '<u8'),
      ('measurement_uid', '<u4'),
      ('scan_counter', '<u4'),
      ('acquisition_time_stamp', '<u4'),
      ('physiology_time_stamp', '<u4', (3,)),
      ('number_of_samples', '<u2'),
      ('available_channels', '<u2'),
      ('active_channels', '<u2'),
      ('channel_mask', '<u8', (16,)),
      ('discard_pre', '<u2'),
      ('discard_post', '<u2'),
      ('center_sample', '<u2'),
      ('encoding_space_ref', '<u2'),
      ('trajectory_dimensions', '<u2'),
      ('sample_time_us', '<f4'),
      ('position', '<f4', (3,)),
      ('read_dir', '<f4', (3,)),
      ('phase_dir', '<f4', (3,)),
      ('slice_dir', '<f4', (3,)),
      ('patient_table_position', '<f4', (3,)),
      ('idx', encoding_counters_dtype),
      ('user_int', '<i4', (8,)),
      ('user_float', '<f4', (8,))])

acquisition_dtype = np.dtype(
    [('head', acquisition_header_dtype),
     ('traj', h5py.special_dtype(vlen=np.dtype('float32'))),
     ('data', h5py.special_dtype(vlen=np.dtype('float32')))])

image_header_dtype = np.dtype(
    [('version', '<u2'),
     ('data_type', '<u2'),
     ('flags', '<u8'),
     ('measurement_uid', '<u4'),
     ('matrix_size', '<u2', (3,)),
     ('field_of_view', '<f4', (3,)),
     ('channels', '<u2'),
     ('position', '<f4', (3,)),
     ('read_dir', '<f4', (3,)),
     ('phase_dir', '<f4', (3,)),
     ('slice_dir', '<f4', (3,)),
     ('patient_table_position', '<f4', (3,)),
     ('average', '<u2'),
     ('slice', '<u2'),
     ('contrast', '<u2'),
     ('phase', '<u2'),
     ('repetition', '<u2'),
     ('set', '<u2'),
     ('acquisition_time_stamp', '<u4'),
     ('physiology_time_stamp', '<u4', (3,)),
     ('image_type', '<u2'),
     ('image_index', '<u2'),
     ('image_series_index', '<u2'),
     ('user_int', '<i4', (8,)),
     ('user_float', '<f4', (8,)),
     ('attribute_string_len', '<u4')])


waveform_header_dtype = np.dtype(
    [('version', '<u2'),
     ('flags', '<u8'),
     ('measurement_uid', '<u4'),
     ('scan_counter', '<u4'),
     ('time_stamp', '<u4'),
     ('number_of_samples', '<u2'),
     ('channels', '<u2'),
     ('sample_time_us', '<f4'),
     ('waveform_id', '<u2')],align=True)

waveform_dtype = np.dtype(
    [('head', waveform_header_dtype),
     ('data', h5py.special_dtype(vlen=np.dtype('uint32')))])


# hdf5 data type lookup function
def get_hdf5type(val):
    if val == DATATYPE_USHORT:
        return np.dtype('<u2')
    elif val == DATATYPE_SHORT:
        return np.dtype('<i2')
    elif val == DATATYPE_UINT:
        return np.dtype('<u4')
    elif val == DATATYPE_INT:
        return np.dtype('<i4')
    elif val == DATATYPE_FLOAT:
        return np.dtype('<f4')
    elif val == DATATYPE_DOUBLE:
        return np.dtype('<f8')
    elif val == DATATYPE_CXFLOAT:
        return np.dtype([('real','<f4'),('imag','<f4')])
    elif val == DATATYPE_CXDOUBLE:
        return np.dtype([('real','<f8'),('imag','<f8')])
    else:
        raise TypeError("Unknown data type.")


def get_arrayhdf5type(val):
    if val == np.uint16:
        return np.dtype('<u2')
    elif val == np.int16:
        return np.dtype('<i2')
    elif val == np.uint32:
        return np.dtype('<u4')
    elif val == np.int32:
        return np.dtype('<i4')
    elif val == np.float32:
        return np.dtype('<f4')
    elif val == np.float64:
        return np.dtype('<f8')
    elif val == np.complex64:
        return np.dtype([('real','<f4'),('imag','<f4')])
    elif val == np.complex128:
        return np.dtype([('real','<f8'),('imag','<f8')])
    else:
        raise TypeError("Unsupported data type.")    
    

def fileinfo(fname):
    with h5py.File(fname, 'r') as f:
        return list(f.keys())


class Dataset(object):
    def __init__(self, filename, dataset_name="dataset", create_if_needed=True):
        # Open the file
        if create_if_needed:
            self._file = h5py.File(filename, 'a')
        else:
            self._file = h5py.File(filename, 'r+')

        self._dataset_name = dataset_name

    def __del__(self):
        try:
            self.close()
        except:
            pass
        
    @property
    def _dataset(self):
        if self._dataset_name not in self._file:
            raise LookupError("Dataset not found in the hdf5 file.")
        return self._file[self._dataset_name]

    def list(self):
        return self._dataset.keys()
    
    def close(self):
        self._file.close()

    def read_xml_header(self):
        if 'xml' not in self._dataset:
            raise LookupError("XML header not found in the dataset.")
        return self._dataset['xml'][0]

    def write_xml_header(self,xmlstring):
        # create the dataset if needed
        self._file.require_group(self._dataset_name)

        dset = self._dataset.require_dataset('xml',shape=(1,), dtype=h5py.special_dtype(vlen=bytes))
        dset[0] = xmlstring

    def number_of_acquisitions(self):
        if 'data' not in self._dataset:
            raise LookupError("Acquisition data not found in the dataset.")
        return self._dataset['data'].size

    def read_acquisition(self, acqnum):
        if 'data' not in self._dataset:
            raise LookupError("Acquisition data not found in the dataset.")
        
        # create an acquisition
        # and fill with the header for this acquisition
        acq = ismrmrd.Acquisition(self._dataset['data'][acqnum]['head'])

        # copy the data as complex float
        acq.data[:] = self._dataset['data'][acqnum]['data'].view(np.complex64).reshape((acq.active_channels, acq.number_of_samples))[:]

        # copy the trajectory as float
        if acq.traj.size>0:
            acq.traj[:] = self._dataset['data'][acqnum]['traj'].reshape((acq.number_of_samples,acq.trajectory_dimensions))[:]

        return acq

    def append_acquisition(self, acq):
        # create the dataset if needed
        self._file.require_group(self._dataset_name)
        
        # extend by 1
        if 'data' in self._dataset:
            acqnum = self._dataset['data'].shape[0]
            self._dataset['data'].resize(acqnum+1,axis=0)
        else:
            self._dataset.create_dataset("data", (1,), maxshape=(None,), dtype=acquisition_dtype)
            acqnum = 0

        # create an empty hdf5 acquisition and fill it
        h5acq = np.empty((1,), dtype=acquisition_dtype)
        # copy the header

        h5acq[0]['head'] = np.frombuffer(acq.getHead(), dtype=acquisition_header_dtype)

        # copy the data as float
        h5acq[0]['data'] = acq.data.view(np.float32).reshape((2*acq.active_channels*acq.number_of_samples,))
        
        # copy the trajectory as float
        h5acq[0]['traj'] = acq.traj.view(np.float32).reshape((acq.number_of_samples*acq.trajectory_dimensions,))

        # put it into the hdf5 file
        self._dataset['data'][acqnum] = h5acq[0]

    def write_acquisition(self,acq,acqnum):

        # create an empty hdf5 acquisition and fill it
        h5acq = np.empty((1,), dtype=acquisition_dtype)
        # copy the header

        h5acq[0]['head'] = np.frombuffer(acq.getHead(), dtype=acquisition_header_dtype)

        # copy the data as float
        h5acq[0]['data'] = acq.data.view(np.float32).reshape((2*acq.active_channels*acq.number_of_samples,))

        # copy the trajectory as float
        h5acq[0]['traj'] = acq.traj.view(np.float32).reshape((acq.number_of_samples*acq.trajectory_dimensions,))

        # put it into the hdf5 file
        self._dataset['data'][acqnum] = h5acq[0]


    def number_of_images(self, impath):
        if impath not in self._dataset:
            raise LookupError("Image data not found in the dataset.")
        return self._dataset[impath]['header'].shape[0]
    
    def read_image(self, impath, imnum):
        if impath not in self._dataset:
            raise LookupError("Image data not found in the dataset.")
        
        # create an image
        # and fill with the header and attribute string for this image
        im = ismrmrd.Image(self._dataset[impath]['header'][imnum], self._dataset[impath]['attributes'][imnum])

        # copy the data
        # ismrmrd complex data is stored as pairs named real and imag
        # TODO do we need to store and reset or the config local to the module?
        cplxcfg = h5py.get_config().complex_names
        h5py.get_config().complex_names = ('real','imag')
        im.data[:] = self._dataset[impath]['data'][imnum]
        h5py.get_config().complex_names = cplxcfg

        return im
    
    def append_image(self, impath, im):
        # create the dataset if needed
        self._file.require_group(self._dataset_name)

        # create the image if needed
        self._dataset.require_group(impath)
        
        # extend by 1
        if 'header' in self._dataset[impath]:
            imnum = self._dataset[impath]['header'].shape[0]
            # check that they are all equal
            self._dataset[impath]['header'].resize(imnum+1,axis=0)
            self._dataset[impath]['attributes'].resize(imnum+1,axis=0)
            self._dataset[impath]['data'].resize(imnum+1,axis=0)
        else:
            self._dataset[impath].create_dataset("header", (1,), maxshape=(None,), dtype=image_header_dtype)
            self._dataset[impath].create_dataset("attributes", (1,), maxshape=(None,), dtype=h5py.special_dtype(vlen=str))
            self._dataset[impath].create_dataset("data", (1,im.data.shape[0],im.data.shape[1],im.data.shape[2],im.data.shape[3]),
                                                maxshape=(None,im.data.shape[0],im.data.shape[1],im.data.shape[2],im.data.shape[3]), dtype=get_hdf5type(im.data_type))
            imnum = 0

        self._dataset[impath]['header'][imnum] = np.frombuffer(im.getHead(), dtype=image_header_dtype)
        # put the attribute string
        self._dataset[impath]['attributes'][imnum] = im.attribute_string
        # put the data
        self._dataset[impath]['data'][imnum] = im.data.view(dtype=get_hdf5type(im.data_type))

    def number_of_arrays(self, arrpath):
        if arrpath not in self._dataset:
            raise LookupError("Array data not found in the dataset.")
        return self._dataset[arrpath].shape[0]
    
    def read_array(self, arrpath, arrnum):
        if arrpath not in self._dataset:
            raise LookupError("Array data not found in the dataset.")
        
        # ismrmrd complex data is stored as pairs named real and imag
        # TODO do we need to store and reset or the config local to the module?
        cplxcfg = h5py.get_config().complex_names
        h5py.get_config().complex_names = ('real','imag')
        arr = np.copy(self._dataset[arrpath][arrnum])
        h5py.get_config().complex_names = cplxcfg

        return arr
    
    def append_array(self, arrpath, arr):
        # create the dataset if needed
        self._file.require_group(self._dataset_name)

        # extend by 1
        if arrpath in self._dataset:
            arrnum = self._dataset[arrpath].shape[0]
            self._dataset[arrpath].resize(arrnum+1,axis=0)
        else:
            shape = list(arr.shape)
            shape.insert(0,1)            
            shape = tuple(shape)
            maxshape = list(arr.shape)
            maxshape.insert(0,None)
            maxshape = tuple(maxshape)
            self._dataset.create_dataset(arrpath, shape, maxshape=maxshape, dtype=get_arrayhdf5type(arr.dtype))
            arrnum = 0
        
        # put the data
        self._dataset[arrpath][arrnum] = arr.view(dtype=get_arrayhdf5type(arr.dtype))

    def number_of_waveforms(self):
        if 'waveforms' not in self._dataset:
            raise LookupError("Acquisition data not found in the dataset.")
        return self._dataset['waveforms'].size

    def read_waveform(self, wavnum):
        if 'waveforms' not in self._dataset:
            raise LookupError("Acquisition data not found in the dataset.")

        # create a Waveform
        # and fill with the header for this waveform
        # HDF5 does not guarantee data alignment, so we must copy into a numpy array. *sigh*
        # We start with an array of zeros to avoid garbage in the padding bytes.
        header_array = np.zeros((1, ), dtype=waveform_header_dtype)
        header_array[0] = self._dataset['waveforms'][wavnum]['head']

        wav = ismrmrd.Waveform(header_array)

        # copy the data as uint32
        wav.data[:] = self._dataset['waveforms'][wavnum]['data'].view(np.uint32).reshape((wav.channels, wav.number_of_samples))[:]

        return wav

    def append_waveform(self, wav):
        # create the dataset if needed
        self._file.require_group(self._dataset_name)

        if 'waveforms' in self._dataset:
            wavnum = self._dataset['waveforms'].shape[0]
            self._dataset['waveforms'].resize(wavnum+1, axis=0)
        else:
            self._dataset.create_dataset("waveforms", (1,), maxshape=(None,), dtype=waveform_dtype)
            wavnum = 0

        np.frombuffer(wav.getHead(), dtype=waveform_header_dtype)

        # create an empty hdf5 acquisition and fill it
        h5wav = np.empty((1,),dtype=waveform_dtype)
        # copy the header

        h5wav[0]['head'] = np.frombuffer(wav.getHead(), dtype=waveform_header_dtype)

        # copy the data as float
        h5wav[0]['data'] = wav.data.view(np.uint32).reshape(( wav.channels * wav.number_of_samples,))

        # put it into the hdf5 file
        self._dataset['waveforms'][wavnum] = h5wav[0]



import ctypes
import numpy as np
import copy
import io

from .flags import FlagsMixin
from .equality import EqualityMixin


class WaveformHeader(FlagsMixin, EqualityMixin, ctypes.Structure):
    _pack_ = 8
    _fields_ = [("version", ctypes.c_uint16),
                ("flags", ctypes.c_uint64),
                ("measurement_uid", ctypes.c_uint32),
                ("scan_counter", ctypes.c_uint32),
                ("time_stamp", ctypes.c_uint32),
                ("number_of_samples", ctypes.c_uint16),
                ("channels", ctypes.c_uint16),
                ("sample_time_us", ctypes.c_float),
                ("waveform_id", ctypes.c_uint16)]

    def __str__(self):
        retstr = ''
        for field_name, field_type in self._fields_:
            if field_name == "__alignment_bits":
                continue
            var = getattr(self, field_name)
            if hasattr(var, '_length_'):
                retstr += '%s: %s\n' % (field_name, ', '.join((str(v) for v in var)))
            else:
                retstr += '%s: %s\n' % (field_name, var)
        return retstr


class Waveform(FlagsMixin):
    __readonly = ('number_of_samples', 'channels')

    @staticmethod
    def deserialize_from(read_exactly):

        header_bytes = read_exactly(ctypes.sizeof(WaveformHeader))
        waveform = Waveform(header_bytes)

        data_bytes = read_exactly(waveform.channels *
                                  waveform.number_of_samples *
                                  ctypes.sizeof(ctypes.c_uint32))

        waveform.data.ravel()[:] = np.frombuffer(data_bytes, dtype=np.uint32)

        return waveform

    def serialize_into(self, write):
        write(self.__head)
        write(self.__data.tobytes())

    @staticmethod
    def from_bytes(bytelike):
        with io.BytesIO(bytelike) as stream:
            return Waveform.deserialize_from(stream.read)

    def to_bytes(self):
        with io.BytesIO() as stream:
            self.serialize_into(stream.write)
            return stream.getvalue()


    @staticmethod
    def from_array(data, **kwargs):

        channels, nsamples = data.shape

        array_data = {
            'version': 1,
            'channels': channels,
            'number_of_samples': nsamples
        }

        header = WaveformHeader()

        properties = dict(array_data, **kwargs)

        for field in properties:
            setattr(header, field, properties.get(field))

        waveform = Waveform(header)
        waveform.data[:] = data

        return waveform

    def __init__(self, head=None, data=None):
        if head is None:
            self.__head = WaveformHeader()
            self.__data = np.empty(shape=(1, 0), dtype=np.uint32)
        else:
            self.__head = WaveformHeader.from_buffer_copy(head)
            self.__data = np.empty(shape=(self.__head.channels, self.__head.number_of_samples), dtype=np.uint32)

            if data is not None:
                self.data[:] = data.reshape((self.__head.channels,self.__head.number_of_samples), order="C")

        for (field, type) in self.__head._fields_:
            try:
                g = '__get_' + field
                s = '__set_' + field
                setattr(Waveform, g, self.__getter(field))
                setattr(Waveform, s, self.__setter(field))
                p = property(getattr(Waveform, g), getattr(Waveform, s))
                setattr(Waveform, field, p)
            except TypeError:
                # e.g. if key is an `int`, skip it
                pass

    def __getter(self, name):
        if name in self.__readonly:
            def fn(self):
                return copy.copy(self.__head.__getattribute__(name))
        else:
            def fn(self):
                return self.__head.__getattribute__(name)
        return fn

    def __setter(self, name):
        if name in self.__readonly:
            def fn(self, val):
                raise AttributeError(name+" is read-only. Use resize instead.")
        else:
            def fn(self, val):
                self.__head.__setattr__(name, val)

        return fn

    def resize(self, number_of_samples=0, channels=1):
        self.__data = np.resize(self.__data, (channels, number_of_samples))
        self.__head.number_of_samples = number_of_samples
        self.__head.channels = channels

    def getHead(self):
        return copy.deepcopy(self.__head)

    def setHead(self, hdr):
        self.__head = self.__head.__class__.from_buffer_copy(hdr)
        self.resize(self.__head.number_of_samples, self.__head.active_channels)

    @property
    def data(self):
        return self.__data.view()

    def __str__(self):
        return "Header:\n {}\nData:\n {}\n".format(self.__head, self.__data)

    def __eq__(self, other):
        if not isinstance(other, Waveform):
            return False

        return all([
            self.__head == other.__head,
            np.array_equal(self.__data, other.__data)
        ])



# Constants
USER_INTS = 8
USER_FLOATS = 8
PHYS_STAMPS = 3
CHANNEL_MASKS = 16
NDARRAY_MAXDIM = 7
POSITION_LENGTH = 3
DIRECTION_LENGTH = 3

# Data types
DATATYPE_USHORT   = 1 # corresponds to uint16_t
DATATYPE_SHORT    = 2 # corresponds to int16_t
DATATYPE_UINT     = 3 # corresponds to uint32_t
DATATYPE_INT      = 4 # corresponds to int32_t
DATATYPE_FLOAT    = 5 # corresponds to float
DATATYPE_DOUBLE   = 6 # corresponds to double
DATATYPE_CXFLOAT  = 7 # corresponds to complex float
DATATYPE_CXDOUBLE = 8 # corresponds to complex double

# Acquisition flags
ACQ_FIRST_IN_ENCODE_STEP1               =  1
ACQ_LAST_IN_ENCODE_STEP1                =  2
ACQ_FIRST_IN_ENCODE_STEP2               =  3
ACQ_LAST_IN_ENCODE_STEP2                =  4
ACQ_FIRST_IN_AVERAGE                    =  5
ACQ_LAST_IN_AVERAGE                     =  6
ACQ_FIRST_IN_SLICE                      =  7
ACQ_LAST_IN_SLICE                       =  8
ACQ_FIRST_IN_CONTRAST                   =  9
ACQ_LAST_IN_CONTRAST                    = 10
ACQ_FIRST_IN_PHASE                      = 11
ACQ_LAST_IN_PHASE                       = 12
ACQ_FIRST_IN_REPETITION                 = 13
ACQ_LAST_IN_REPETITION                  = 14
ACQ_FIRST_IN_SET                        = 15
ACQ_LAST_IN_SET                         = 16
ACQ_FIRST_IN_SEGMENT                    = 17
ACQ_LAST_IN_SEGMENT                     = 18
ACQ_IS_NOISE_MEASUREMENT                = 19
ACQ_IS_PARALLEL_CALIBRATION             = 20
ACQ_IS_PARALLEL_CALIBRATION_AND_IMAGING = 21
ACQ_IS_REVERSE                          = 22
ACQ_IS_NAVIGATION_DATA                  = 23
ACQ_IS_PHASECORR_DATA                   = 24
ACQ_LAST_IN_MEASUREMENT                 = 25
ACQ_IS_HPFEEDBACK_DATA                  = 26
ACQ_IS_DUMMYSCAN_DATA                   = 27
ACQ_IS_RTFEEDBACK_DATA                  = 28
ACQ_IS_SURFACECOILCORRECTIONSCAN_DATA   = 29
ACQ_IS_PHASE_STABILIZATION_REFERENCE    = 30
ACQ_IS_PHASE_STABILIZATION              = 31
ACQ_COMPRESSION1                        = 53
ACQ_COMPRESSION2                        = 54
ACQ_COMPRESSION3                        = 55
ACQ_COMPRESSION4                        = 56
ACQ_USER1                               = 57
ACQ_USER2                               = 58
ACQ_USER3                               = 59
ACQ_USER4                               = 60
ACQ_USER5                               = 61
ACQ_USER6                               = 62
ACQ_USER7                               = 63
ACQ_USER8                               = 64

# Image types
IMTYPE_MAGNITUDE = 1
IMTYPE_PHASE     = 2
IMTYPE_REAL      = 3
IMTYPE_IMAG      = 4
IMTYPE_COMPLEX   = 5

# Image flags
IMAGE_IS_NAVIGATION_DATA  =  1
IMAGE_FIRST_IN_AVERAGE    =  5
IMAGE_LAST_IN_AVERAGE     =  6
IMAGE_FIRST_IN_SLICE      =  7
IMAGE_LAST_IN_SLICE       =  8
IMAGE_FIRST_IN_CONTRAST   =  9
IMAGE_LAST_IN_CONTRAST    = 10
IMAGE_FIRST_IN_PHASE      = 11
IMAGE_LAST_IN_PHASE       = 12
IMAGE_FIRST_IN_REPETITION = 13
IMAGE_LAST_IN_REPETITION  = 14
IMAGE_FIRST_IN_SET        = 15
IMAGE_LAST_IN_SET         = 16
IMAGE_USER1               = 57
IMAGE_USER2               = 58
IMAGE_USER3               = 59
IMAGE_USER4               = 60
IMAGE_USER5               = 61
IMAGE_USER6               = 62
IMAGE_USER7               = 63
IMAGE_USER8               = 64





class FlagsMixin(object):

    def clearAllFlags(self):
        self.clear_all_flags()

    def isFlagSet(self, flag):
        return self.is_flag_set(flag)

    def setFlag(self, flag):
        self.set_flag(flag)

    def clearFlag(self, flag):
        self.clear_flag(flag)

    def clear_all_flags(self):
        self.flags = 0

    def is_flag_set(self, flag):
        return bool(self.flags & (1 << (flag - 1)))

    def set_flag(self, flag):
        self.flags |= (1 << (flag - 1))

    def clear_flag(self, flag):
        self.flags &= ~(1 << (flag - 1))




from .ismrmrdschema import ismrmrdHeader

from xsdata.formats.dataclass.parsers import XmlParser
from xsdata.formats.dataclass.parsers.config import ParserConfig

from xsdata.formats.dataclass.serializers import XmlSerializer
from xsdata.formats.dataclass.serializers.config import SerializerConfig
import xml.dom.minidom as md

def CreateFromDocument(document):
    parser = XmlParser(config=ParserConfig(fail_on_unknown_properties=True))
    if isinstance(document,str):
        return parser.from_string(document,ismrmrdHeader)
    return parser.from_bytes(document,ismrmrdHeader)


def ToXML(header: ismrmrdHeader , encoding='ascii'):
    config = SerializerConfig(encoding=encoding,pretty_print=True)
    serializer = XmlSerializer(config)
    return serializer.render(header,ns_map={"":"http://www.ismrm.org/ISMRMRD"})

def ToDOM(header: ismrmrdHeader):
    return md.parseString(ToXML(header))


ismrmrdHeader.toXML = ToXML
ismrmrdHeader.toDOM = ToDOM


import h5py
import numpy

from .hdf5 import *
from .acquisition import *
from .waveform import *
from .image import *
from .xsd import ToXML


class DataWrapper:

    def __init__(self, data):
        self.data = data

    def __len__(self):
        return self.data.size

    def __iter__(self):
        for raw in self.data:
            yield self.from_numpy(raw)

    def __getitem__(self, key):
        if isinstance(key, slice):
            return [self.from_numpy(raw) for raw in self.data[key]]
        else:
            return self.from_numpy(self.data[key])

    def __setitem__(self, key, value):
        try:
            iterable = [self.to_numpy(v) for v in value]
        except TypeError:
            iterable = [self.to_numpy(value)]

        self.data[key] = numpy.array(iterable, dtype=self.datatype)

    def __repr__(self):
        return type(self).__name__ + " containing " + self.data.__repr__()

    @classmethod
    def from_numpy(cls, raw):
        raise NotImplemented()

    @classmethod
    def to_numpy(cls, item):
        raise NotImplemented()

    datatype = None


class Acquisitions(DataWrapper):

    def __init__(self, data):
        super().__init__(data)

    @classmethod
    def from_numpy(cls, raw):
        acquisition = Acquisition(raw['head'])

        acquisition.data[:] = raw['data'].view(np.complex64).reshape(
            (acquisition.active_channels,
             acquisition.number_of_samples)
        )[:]

        acquisition.traj[:] = raw['traj'].reshape(
            (acquisition.number_of_samples,
             acquisition.trajectory_dimensions)
        )[:]

        return acquisition

    @classmethod
    def to_numpy(cls, acq):
        return (
            np.frombuffer(acq.getHead(), dtype=acquisition_header_dtype),
            acq.traj.view(np.float32).reshape((acq.number_of_samples * acq.trajectory_dimensions,)),
            acq.data.view(np.float32).reshape((2 * acq.active_channels * acq.number_of_samples,))
        )

    datatype = acquisition_dtype


class Waveforms(DataWrapper):

    def __init__(self, data):
        super().__init__(data)

    @classmethod
    def from_numpy(cls, raw):
        # This lovely roundabout way of reading a waveform header is due largely to h5's handling
        # of padding and alignment. Nu guarantees are given, so we need create a structured array
        # with a header to have the contents filled in correctly. We start with an array of
        # zeroes to avoid garbage in the padding bytes.
        header_array = numpy.zeros((1,), dtype=waveform_header_dtype)
        header_array[0] = raw['head']

        waveform = Waveform(header_array)
        waveform.data[:] = raw['data'].view(np.uint32).reshape(
            (waveform.channels,
             waveform.number_of_samples)
        )[:]

        return waveform

    @classmethod
    def to_numpy(cls, wav):
        return (
            np.frombuffer(wav.getHead(), dtype=waveform_header_dtype),
            wav.data.view(np.uint32).reshape((wav.channels * wav.number_of_samples,))
        )

    datatype = waveform_dtype


class Images:

    def __init__(self, data, headers, attributes):
        self.data = data
        self.headers = headers
        self.attributes = attributes

    def __len__(self):
        return self.headers.size

    def __iter__(self):
        for data, header, attributes in zip(self.data, self.headers, self.attributes):
            yield self.from_numpy(header, data, attributes)

    def __getitem__(self, key):
        if isinstance(key, slice):
            return [self.from_numpy(*raw) for raw in zip(self.headers[key], self.data[key], self.attributes[key])]
        else:
            return self.from_numpy(
                self.headers[key],
                self.data[key],
                self.attributes[key]
            )

    def __setitem__(self, key, value):
        try:
            iterable = [self.to_numpy(v) for v in value]
        except TypeError:
            iterable = [self.to_numpy(value)]

        self.headers[key] = numpy.stack([header for header, _, __ in iterable])
        self.data[key] = numpy.stack([data for _, data, __ in iterable])
        self.attributes[key] = numpy.stack([attributes for _, __, attributes in iterable])

    @classmethod
    def from_numpy(cls, header, data, attributes):
        image = Image(header, attributes.decode('ascii', 'strict'))
        image.data[:] = data.astype(image.data.dtype)
        return image

    @classmethod
    def to_numpy(cls, image):
        header = np.frombuffer(image.getHead(), dtype=image_header_dtype)
        data = image.data
        attributes = image.attribute_string.encode(encoding='ascii', errors='strict')

        return header, data, attributes

    datatype = None


class Folder:
    def __init__(self, contents):
        self._contents = contents

    def __getitem__(self, key):
        if key in self._contents:
            return Container(self._contents[key])
        return self.__missing__(key)

    def __delitem__(self, key):
        if key in self._contents:
            del self._contents[key]

    def __missing__(self, key):
        return Container(self._contents.require_group(key))

    def __contains__(self, key):
        return key in self._contents

    def __iter__(self):
        for key, item in self._contents.items():
            if isinstance(item, h5py.Group):
                yield key

    def __str__(self):
        return str([key for key in self._contents])

    def __visit(self, callback, path):

        callback(self, path)

        for key in self:
            child = self[key]
            child.__visit(callback, path + '/' + key)

    def visit(self, callback):
        for key in self:
            child = self[key]
            child.__visit(callback, key)

    def find_images(self):

        images = set()

        def find_images(node, path):
            if node.has_images():
                images.add(path)

        self.visit(find_images)
        return images

    def find_data(self):

        data = set()

        def find_data(node, path):
            if node.has_acquisitions() or node.has_waveforms():
                data.add(path)

        self.visit(find_data)
        return data

    def keys(self):
        return self._contents.keys()


class Container(Folder):

    def __init__(self, contents):
        super(Container, self).__init__(contents)

    def __get_acquisitions(self):
        if not self.has_acquisitions():
            return None
        data = self._contents.get('data')
        return Acquisitions(data)

    def __set_acquisitions(self, acquisitions):

        if self.has_images():
            raise TypeError("Cannot add acquisitions when images are present.")

        buffer = numpy.array([Acquisitions.to_numpy(a) for a in acquisitions], dtype=acquisition_dtype)

        self.__del_acquisitions()
        self._contents['data'] = buffer

    def __del_acquisitions(self):
        if 'data' in self._contents:
            del self._contents['data']

    acquisitions = property(__get_acquisitions, __set_acquisitions, __del_acquisitions)

    def __get_waveforms(self):
        if not self.has_waveforms():
            return None
        data = self._contents.get('waveforms')
        return Waveforms(data)

    def __set_waveforms(self, waveforms):

        if self.has_images():
            raise TypeError("Cannot add waveforms when images are present.")

        converter = Waveforms(None)
        buffer = numpy.array([converter.to_numpy(w) for w in waveforms], dtype=waveform_dtype)

        self.__del_waveforms()
        self._contents['waveforms'] = buffer

    def __del_waveforms(self):
        if 'waveforms' in self._contents:
            del self._contents['waveforms']

    waveforms = property(__get_waveforms, __set_waveforms, __del_waveforms)

    def __get_images(self):
        if not self.has_images():
            return None

        return Images(
            self._contents.get('data'),
            self._contents.get('header'),
            self._contents.get('attributes')
        )

    def __set_images(self, images):

        if self.has_data():
            raise TypeError("Cannot add images when data is present.")

        images = list(images)

        data = numpy.stack([image.data for image in images])
        headers = numpy.stack([np.frombuffer(image.getHead(), dtype=image_header_dtype) for image in images])
        attributes = numpy.stack([image.attribute_string for image in images])

        self.__del_images()
        self._contents.create_dataset('data', data=data)
        self._contents.create_dataset('header', data=headers)
        self._contents.create_dataset('attributes', shape=attributes.shape, dtype=h5py.special_dtype(vlen=bytes))
        self._contents['attributes'][:] = attributes

    def __del_images(self):
        for key in ['header', 'data', 'attributes']:
            if key in self._contents:
                del self._contents[key]

    images = property(__get_images, __set_images, __del_images)

    def __get_header(self):
        if not self.has_header():
            return None
        return ismrmrd.xsd.CreateFromDocument(self._contents['xml'][0])

    def __set_header(self, header):
        self.__del_header()
        self._contents.create_dataset('xml', shape=(1,), dtype=h5py.special_dtype(vlen=bytes))
        self._contents['xml'][0] = ToXML(header)

    def __del_header(self):
        if 'xml' in self._contents:
            del self._contents['xml']

    header = property(__get_header, __set_header, __del_header)

    def available(self):

        def header():
            if self.has_header():
                return 'header'

        def acquisitions():
            if self.has_acquisitions():
                return 'acquisitions'

        def waveforms():
            if self.has_waveforms():
                return 'waveforms'

        def images():
            if self.has_images():
                return 'images'

        return list(filter(bool, [header(), acquisitions(), waveforms(), images()]))

    def has_header(self):
        return 'xml' in self._contents

    def has_images(self):
        return all((key in self._contents for key in ['data', 'header', 'attributes']))

    def has_data(self):
        return any((self.has_acquisitions(), self.has_waveforms()))

    def has_waveforms(self):
        return 'waveforms' in self._contents

    def has_acquisitions(self):
        return 'data' in self._contents and not self.has_images()


class File(Folder):

    def __init__(self, filename, mode='a'):
        self.__file = h5py.File(filename, mode)
        super().__init__(self.__file)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.__file.close()



from ismrmrd.meta import Meta
import itertools
import ctypes
import numpy as np
import copy
import io
import warnings

warnings.simplefilter('default')

from .acquisition import Acquisition
from .flags import FlagsMixin
from .equality import EqualityMixin
from .constants import *

dtype_mapping = {
    DATATYPE_USHORT: np.dtype('uint16'),
    DATATYPE_SHORT: np.dtype('int16'),
    DATATYPE_UINT: np.dtype('uint32'),
    DATATYPE_INT: np.dtype('int'),
    DATATYPE_FLOAT: np.dtype('float32'),
    DATATYPE_DOUBLE: np.dtype('float64'),
    DATATYPE_CXFLOAT: np.dtype('complex64'),
    DATATYPE_CXDOUBLE: np.dtype('complex128')
}
inverse_dtype_mapping = {dtype_mapping.get(k): k for k in dtype_mapping}


def get_dtype_from_data_type(val):
    dtype = dtype_mapping.get(val)
    if dtype is None:
        raise TypeError("Unknown image data type: " + str(val))
    return dtype


def get_data_type_from_dtype(dtype):
    type = inverse_dtype_mapping.get(dtype)
    if type is None:
        raise TypeError("Datatype not supported: " + str(dtype))
    return type


# Image Header
class ImageHeader(FlagsMixin, EqualityMixin, ctypes.Structure):
    _pack_ = 2
    _fields_ = [("version", ctypes.c_uint16),
                ("data_type", ctypes.c_uint16),
                ("flags", ctypes.c_uint64),
                ("measurement_uid", ctypes.c_uint32),
                ("matrix_size", ctypes.c_uint16 * POSITION_LENGTH),
                ("field_of_view", ctypes.c_float * POSITION_LENGTH),
                ("channels", ctypes.c_uint16),
                ("position", ctypes.c_float * POSITION_LENGTH),
                ("read_dir", ctypes.c_float * DIRECTION_LENGTH),
                ("phase_dir", ctypes.c_float * DIRECTION_LENGTH),
                ("slice_dir", ctypes.c_float * DIRECTION_LENGTH),
                ("patient_table_position", ctypes.c_float * POSITION_LENGTH),
                ("average", ctypes.c_uint16),
                ("slice", ctypes.c_uint16),
                ("contrast", ctypes.c_uint16),
                ("phase", ctypes.c_uint16),
                ("repetition", ctypes.c_uint16),
                ("set", ctypes.c_uint16),
                ("acquisition_time_stamp", ctypes.c_uint32),
                ("physiology_time_stamp", ctypes.c_uint32 * PHYS_STAMPS),
                ("image_type", ctypes.c_uint16),
                ("image_index", ctypes.c_uint16),
                ("image_series_index", ctypes.c_uint16),
                ("user_int", ctypes.c_int32 * USER_INTS),
                ("user_float", ctypes.c_float * USER_FLOATS),
                ("attribute_string_len", ctypes.c_uint32), ]

    @staticmethod
    def from_acquisition(acquisition, **kwargs):

        header = ImageHeader()

        # The value of these fields is copied over from the acquisition header.
        acquisition_fields = [
            'version',
            'measurement_uid',
            'position',
            'read_dir',
            'phase_dir',
            'slice_dir',
            'patient_table_position',
            'acquisition_time_stamp',
            'physiology_time_stamp',
            'user_int',
            'user_float'
        ]

        for field in acquisition_fields:
            setattr(header, field, getattr(acquisition, field))

        idx_fields = ['average', 'slice', 'contrast', 'phase', 'repetition', 'set']
        for field in idx_fields:
            setattr(header, field, getattr(acquisition.idx, field))

        for field in kwargs:
            setattr(header, field, kwargs.get(field))

        return header

    def __str__(self):
        retstr = ''
        for field_name, field_type in self._fields_:
            var = getattr(self, field_name)
            if hasattr(var, '_length_'):
                retstr += '%s: %s\n' % (field_name, ', '.join((str(v) for v in var)))
            else:
                retstr += '%s: %s\n' % (field_name, var)
        return retstr

    def __repr__(self):
        retstr = 'ImageHeader('
        for field_name, field_type in self._fields_:
            var = getattr(self, field_name)
            if hasattr(var, '_length_'):
                retstr += '%s: (%s), ' % (field_name, ', '.join((str(v) for v in var)))
            else:
                retstr += '%s: %s, ' % (field_name, var)
        retstr += ')'
        return retstr


# Image class
class Image(FlagsMixin):
    __readonly = ('data_type', 'matrix_size', 'channels')
    __ignore = ('matrix_size', 'attribute_string_len')

    @staticmethod
    def deserialize_from(read_exactly):

        header_bytes = read_exactly(ctypes.sizeof(ImageHeader))
        attribute_length_bytes = read_exactly(ctypes.sizeof(ctypes.c_uint64))
        attribute_length = ctypes.c_uint64.from_buffer_copy(attribute_length_bytes)
        attribute_bytes = read_exactly(attribute_length.value).rstrip(b'\0')

        image = Image(header_bytes, attribute_bytes.decode('utf-8'))

        def calculate_number_of_entries(nchannels, xs, ys, zs):
            return nchannels * xs * ys * zs

        nentries = calculate_number_of_entries(image.channels, *image.matrix_size)
        nbytes = nentries * get_dtype_from_data_type(image.data_type).itemsize

        data_bytes = read_exactly(nbytes)

        image.data.ravel()[:] = np.frombuffer(data_bytes, dtype=get_dtype_from_data_type(image.data_type))

        return image

    def serialize_into(self, write):

        attribute_bytes = self.attribute_string.encode('utf-8')
        self.__head.attribute_string_len = len(attribute_bytes)

        write(self.__head)

        write(ctypes.c_uint64(len(attribute_bytes)))
        write(attribute_bytes)

        write(self.__data.tobytes())

    @staticmethod
    def from_bytes(bytelike):
        with io.BytesIO(bytelike) as stream:
            return Image.deserialize_from(stream.read)

    def to_bytes(self):
        with io.BytesIO() as stream:
            self.serialize_into(stream.write)
            return stream.getvalue()

    @staticmethod
    def from_array(array, acquisition=Acquisition(), transpose=True, **kwargs):

        def input_shape_to_header_format(array):
            def with_defaults(first=1, second=1, third=1, nchannels=1):
                return nchannels, (first, second, third)

            return with_defaults(*reversed(array.shape))

        def header_format_to_resize_shape(nchannels, first, second, third):
            return nchannels, third, second, first

        if transpose:
            warnings.warn(
                "The default behavior of this function is currently column-major which " +
                "is inconsistent with numpy using row-major by default. In a future " +
                "version this will be changed. Please switch to setting transpose in " +
                "this function to false to switch to the new behavior.",
                PendingDeprecationWarning
            )
            array = array.transpose()

        nchannels, matrix_size = input_shape_to_header_format(array)

        image_properties = {
            'version': 1,
            'data_type': get_data_type_from_dtype(array.dtype),
            'channels': nchannels,
            'matrix_size': matrix_size
        }

        header = ImageHeader.from_acquisition(acquisition, **dict(image_properties, **kwargs))
        image = Image(head=header)

        image.data[:] = np.resize(array, header_format_to_resize_shape(nchannels, *matrix_size))

        return image

    def __init__(self, head=None, attribute_string=None, meta=None, data=None):

        if data is not None:
            assert data.ndim <= 4
            data = np.array(data,ndmin=4,copy=False)

        def create_consistent_header(header, data):

            if data is not None:
                header.data_type = inverse_dtype_mapping[data.dtype]
                header.channels = data.shape[0]
                header.matrix_size = (data.shape[3], data.shape[2], data.shape[1])

            return header

        if head is None:
            if data is None:
                data = np.empty((1, 1, 1, 0), dtype=np.complex64)
            self.__head = create_consistent_header(ImageHeader(), data)
        else:
            self.__head = ImageHeader.from_buffer_copy(head)
            if data is None:
                data = np.empty(shape=(self.__head.channels, self.__head.matrix_size[2],
                                              self.__head.matrix_size[1], self.__head.matrix_size[0]),
                                       dtype=get_dtype_from_data_type(self.__head.data_type))
            else:
                self.__head = create_consistent_header(self.__head, data)
        self.__data = data

        if attribute_string is not None:
            if meta is not None:
                raise RuntimeError("Attribute string and meta cannot be set simultaneously.")
            self.__meta = Meta.deserialize(attribute_string)
        elif meta is not None:
            self.__meta = meta
        else:
            self.__meta = Meta()

        for (field, type) in self.__head._fields_:
            if field in self.__ignore:
                continue
            else:
                try:
                    g = '__get_' + field
                    s = '__set_' + field
                    setattr(Image, g, self.__getter(field))
                    setattr(Image, s, self.__setter(field))
                    p = property(getattr(Image, g), getattr(Image, s))
                    setattr(Image, field, p)
                except TypeError:
                    # e.g. if key is an `int`, skip it
                    pass

    def __getter(self, name):
        if name in self.__readonly:
            def fn(self):
                return copy.copy(self.__head.__getattribute__(name))
        else:
            def fn(self):
                return self.__head.__getattribute__(name)
        return fn

    def __setter(self, name):
        if name in self.__readonly:
            def fn(self, val):
                raise AttributeError(name + " is read-only.")
        else:
            def fn(self, val):
                self.__head.__setattr__(name, val)

        return fn

    def getHead(self):
        return copy.deepcopy(self.__head)

    def setHead(self, hdr):
        self.__head = self.__head.__class__.from_buffer_copy(hdr)
        self.setDataType(self.__head.data_type)
        self.resize(self.__head.channels, self.__head.matrix_size[2], self.__head.matrix_size[1],
                    self.__head.matrix_size[0])

    def setDataType(self, val):
        self.__data = self.__data.astype(get_dtype_from_data_type(val))

    def resize(self, nc, nz, ny, nx):
        self.__data = np.resize(self.__data, (nc, nz, ny, nx))

    @property
    def data(self):
        return self.__data.view()

    @property
    def attribute_string(self):
        return self.__meta.serialize()

    @attribute_string.setter
    def attribute_string(self, val):
        self.__meta = Meta.deserialize(val)

    @property
    def meta(self):
        return self.__meta

    @meta.setter
    def meta(self, val):
        if type(val) == Meta:
            self.__meta = val
        elif type(val) == dict:
            self.__meta = Meta()
            self.__meta.update(val)
        else:
            raise RuntimeError("meta must be of type Meta or dict")

    @property
    def matrix_size(self, warn=True):
        if warn:
            warnings.warn(
                "This function currently returns a result that is inconsistent (transposed) " +
                "compared to the matrix_size in the ImageHeader and from .getHead().matrix_size. " +
                "This function will be made consistent in a future version and this message " +
                "will be removed."
            )
        return self.__data.shape[1:4]

    @property
    def attribute_string_len(self):
        return len(self.attribute_string)

    def __str__(self):
        return "Header:\n {}\nAttribute string:\n {}\nData:\n {}\n".format(self.__head, self.attribute_string,
                                                                           self.__data)

    def __repr__(self):
        return f"Image(head={self.__head.__repr__()},meta={self.__meta.__repr__()},data={self.__data.__repr__()})"

    def __eq__(self, other):
        if not isinstance(other, Image):
            return False

        return all([
            self.__head == other.__head,
            np.array_equal(self.__data, other.__data),
            np.array_equal(self.attribute_string, other.attribute_string)
        ])



import os
from setuptools import setup, find_packages
from distutils.command.build import build
from distutils.command.build_py import build_py

import logging
import shutil
from pathlib import Path
import re 

logging.basicConfig()
log_ = logging.getLogger(__name__)

schema_file = os.path.join('schema','ismrmrd.xsd')
config_file = os.path.join('schema','.xsdata.xml')

class my_build_py(build_py):
    def run(self):
        # honor the --dry-run flag
        if not self.dry_run:
            outloc = self.build_lib
            outloc = os.path.join(outloc,'ismrmrd/xsd/ismrmrdschema')

            
            generate_schema(schema_file, config_file, outloc)
                    # distutils uses old-style classes, so no super()
        build_py.run(self)


def fix_init_file(package_name,filepath):

    with open(filepath,'r+') as f:
        text = f.read()
        text = re.sub(f'from {package_name}.ismrmrd', 'from .ismrmrd',text)
        f.seek(0)
        f.write(text)
        f.truncate()




def generate_schema(schema_filename, config_filename, outloc  ):
    def to_uri(filename):
        return Path(filename).absolute().as_uri()

    import sys 
    import subprocess
    
    subpackage_name = 'ismrmrdschema'
    args = [sys.executable,'-m','xsdata', str(schema_filename), '--config',str(config_filename), '--package',subpackage_name]
    subprocess.run(args)
    fix_init_file(subpackage_name,f"{subpackage_name}/__init__.py")
    shutil.rmtree(os.path.join(outloc,subpackage_name),ignore_errors=True)
    shutil.move(subpackage_name,outloc)

this_directory = Path(__file__).parent
long_description = (this_directory / "README").read_text()

setup(
    name='ismrmrd',
    version='1.12.0',
    author='ISMRMRD Developers',
    description='Python implementation of the ISMRMRD',
    license='Public Domain',
    keywords='ismrmrd',
    url='https://ismrmrd.github.io',
    long_description = long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: Public Domain',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: Medical Science Apps.'
        ],
    install_requires=['xsdata>=22.2', 'numpy', 'h5py>=2.3'],
    setup_requires=['nose>=1.0', 'xsdata[cli]>=22.2', 'jinja2 >= 2.11'],
    test_suite='nose.collector',
    cmdclass={'build_py':my_build_py}
)
